﻿Option Explicit On
Option Strict On

Imports Game_PluginAPI_v3
Imports System.IO.MemoryMappedFiles
Imports System.Math
Imports System.Runtime.InteropServices
Public Class GamePlugin
    Implements IPlugin_Game_v3
#Region "    '/// BUILT IN METHODS FOR SIMTOOLS ///"
    '/////////////////////////////////////////////////////
    '/// DO NOT CHANGE - DO NOT CHANGE - DO NOT CHANGE ///
    '/////////////////////////////////////////////////////

    'Output structures
    Private _PluginSettings As New IPlugin_Game_v3.Struct_PluginSettings
    Private _Motion_Outputs As New IPlugin_Game_v3.Struct_MotionOutput
    Private _Dash_Outputs As New IPlugin_Game_v3.Struct_DashOutput
    Private _Vibe_Outputs As New IPlugin_Game_v3.Struct_VibeOutput
    Private _Status_Outputs As New IPlugin_Game_v3.Struct_StatusOutput
    Private Property _OnlineID As String Implements IPlugin_Game_v3.OnlineID  'set what id number

    'Load settings
    Public Sub New()
        _Dash_Outputs.Reset()
        _PluginSettings._PluginVersion = Reflection.Assembly.GetExecutingAssembly.GetName.Version.ToString
        PluginSettings()
    End Sub

    'Output - plugin settings
    Public Function Get_PluginSettings() As IPlugin_Game_v3.Struct_PluginSettings Implements IPlugin_Game_v3.Get_Plugin_Settings
        Return _PluginSettings
    End Function

    'Output - status outputs
    Public Function Get_StatusOutputs() As IPlugin_Game_v3.Struct_StatusOutput Implements IPlugin_Game_v3.Get_Status_Outputs
        Return _Status_Outputs
    End Function

    'Output - motion outputs
    Public Function Get_MotionOutputs() As IPlugin_Game_v3.Struct_MotionOutput Implements IPlugin_Game_v3.Get_Motion_Outputs
        Return _Motion_Outputs
    End Function

    'Output - dash outputs
    Public Function Get_DashOutputs() As IPlugin_Game_v3.Struct_DashOutput Implements IPlugin_Game_v3.Get_Dash_Outputs
        Return _Dash_Outputs
    End Function

    'Output - vibe outputs
    Public Function Get_VibeOutputs() As IPlugin_Game_v3.Struct_VibeOutput Implements IPlugin_Game_v3.Get_Vibe_Outputs
        Return _Vibe_Outputs
    End Function

    'Reset data
    Public Sub PluginReset() Implements IPlugin_Game_v3.PluginReset
        _Motion_Outputs = New IPlugin_Game_v3.Struct_MotionOutput
        _Dash_Outputs.Reset()
        _Vibe_Outputs = New IPlugin_Game_v3.Struct_VibeOutput
    End Sub

    'Launch Shortcut 
    Event LaunchShortcut(ShortcutType As IPlugin_Game_v3.Shortcut_Type, OnlineID As String, Path As String) Implements IPlugin_Game_v3.LaunchShortcut
    Private Sub Create_LaunchShortcut(ShortcutType As IPlugin_Game_v3.Shortcut_Type, Path As String) Implements IPlugin_Game_v3.Create_LaunchShortcut
        RaiseEvent LaunchShortcut(ShortcutType, _OnlineID, Path)
    End Sub

    'Error Logging
    Event ErrorLog(ErrorLevel As IPlugin_Game_v3.Error_Level, Message As String) Implements IPlugin_Game_v3.ErrorLog
    Private Sub Log(ErrorLevel As IPlugin_Game_v3.Error_Level, Message As String) Implements IPlugin_Game_v3.Log
        RaiseEvent ErrorLog(ErrorLevel, Message)
    End Sub

    'Vibe type enum
    Enum VibeType
        NotUsed
        Afterburners
        Chassis
        Collision_LR
        Collision_FB
        Flaps
        Gear_Shift
        Landing_Gear
        Road_Detail
        Rpm
        Turbo
        Turbulence
        Weapons_Bombs
        Weapons_Cannons
        Weapons_Flairs
        Weapons_Guns
        Weapons_Other
    End Enum

    'Platform type enum
    Private Enum PlatformType
        PC
        Console
    End Enum

    'Game origin enum
    Enum GameOrigin
        CD
        Steam
        Epic
        Origin
    End Enum

    'Game type enum
    Enum GameType
        Air
        Land
        Sea
    End Enum

    'Motion type enum
    Enum Motion_Type
        Degree
        Acceleration
    End Enum

    '/////////////////////////////////////////////////////
    '/// DO NOT CHANGE - DO NOT CHANGE - DO NOT CHANGE ///
    '/////////////////////////////////////////////////////
#End Region
    '/////////////////////////////////////////////////////////////////////////////////////////////////////////////
    '///        SimTools GamePlugin - Edit the Setting below to provide support for your favorite game!        ///
    '/////////////////////////////////////////////////////////////////////////////////////////////////////////////

    'Plugin Settings
    Private Sub PluginSettings()
        '//////////////////////////////////////////////// 
        '/// Per Game Settings - Change for Each Game ///
        '////////////////////////////////////////////////
        _PluginSettings._PluginAuthorsName = "value1, yobuddy" 'The plugin authors names (comma delimited)
        _PluginSettings._GameName = "Assetto Corsa" 'GameName - the displayed name in simtools
        _PluginSettings._ProcessName = "AssettoCorsa,acs,acs_x86,acs_pro" 'the Process Name(s) without the (".exe") (comma delimited)
        _PluginSettings._RequiresPatchingPath = False 'do we need a path for patching?
        _PluginSettings._PlatformType = PlatformType.PC 'set the platform type
        _PluginSettings._GameOrigin = GameOrigin.Steam 'set the game origin
        _PluginSettings._GameType = GameType.Land 'set the game type
        _PluginSettings._GameID = "244210" 'set the game ID

        '//////////////////////////////////////////////// 
        '///              Auto Profiling              ///
        '////////////////////////////////////////////////
        '(if set to true, '_Status_Outputs._CurrentVehicle' must get set with the motion data.)
        _PluginSettings._AutoProfiling_Enabled = True 'does the plugin support auto profiling?

        '//////////////////////////////////////////////// 
        '///       Game Provided Crash Variable       ///
        '////////////////////////////////////////////////
        _PluginSettings._GameProvidedCrashed_Enabled = False
        'if set to true, '_Motion_Outputs._HasCrashed' must get set with the motion data below.

        '////////////////////////////////////////////////
        '///      Supported Output for this Game      ///
        '////////////////////////////////////////////////
        _PluginSettings._MotionSupported_Roll = True
        _PluginSettings._MotionSupported_Pitch = True
        _PluginSettings._MotionSupported_Yaw = True
        _PluginSettings._MotionSupported_Heave = True
        _PluginSettings._MotionSupported_Sway = True
        _PluginSettings._MotionSupported_Surge = True
        _PluginSettings._MotionSupported_Extra1 = "Traction Loss" 'Traction Loss LV (Add a name to turn it on)
        _PluginSettings._MotionSupported_Extra2 = "" '("" = off)
        _PluginSettings._MotionSupported_Extra3 = ""
        _PluginSettings._MotionSupported_Extra4 = "Roll Velocity"
        _PluginSettings._MotionSupported_Extra5 = "Pitch Velocity"
        _PluginSettings._MotionSupported_Extra6 = "Yaw Velocity"

        '////////////////////////////////////////////////
        '///     Extra's OutputType for this Game     ///
        '////////////////////////////////////////////////
        _PluginSettings._MotionSupported_Roll_Type = Motion_Type.Degree 'Set the motion type
        _PluginSettings._MotionSupported_Pitch_Type = Motion_Type.Degree
        _PluginSettings._MotionSupported_Yaw_Type = Motion_Type.Degree
        _PluginSettings._MotionSupported_Heave_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Sway_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Surge_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Extra1_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Extra2_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Extra3_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Extra4_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Extra5_Type = Motion_Type.Acceleration
        _PluginSettings._MotionSupported_Extra6_Type = Motion_Type.Acceleration

        '////////////////////////////////////////////////
        '///              Vibe Supported              ///
        '////////////////////////////////////////////////
        _PluginSettings._VibeSupport_VibeEnabled = True
        _PluginSettings._VibeSupported_Vibe01 = VibeType.Rpm
        _PluginSettings._VibeSupported_Vibe02 = VibeType.Gear_Shift
        _PluginSettings._VibeSupported_Vibe03 = VibeType.Collision_LR
        _PluginSettings._VibeSupported_Vibe04 = VibeType.Collision_FB
        _PluginSettings._VibeSupported_Vibe05 = VibeType.Road_Detail
        _PluginSettings._VibeSupported_Vibe06 = VibeType.Chassis
        _PluginSettings._VibeSupported_Vibe07 = VibeType.NotUsed
        _PluginSettings._VibeSupported_Vibe08 = VibeType.NotUsed
        _PluginSettings._VibeSupported_Vibe09 = VibeType.NotUsed
        _PluginSettings._VibeSupported_Vibe10 = VibeType.NotUsed
        _PluginSettings._VibeSupported_Vibe11 = VibeType.NotUsed
        _PluginSettings._VibeSupported_Vibe12 = VibeType.NotUsed

        '////////////////////////////////////////////////
        '///              Dash Supported              ///
        '////////////////////////////////////////////////
        _PluginSettings._DashSupport_DashEnabled = True

        'Speed
        _PluginSettings._DashSupported_Speed_GroundSpeed = True
        _PluginSettings._DashSupported_Speed_AirSpeedIndicator_ASI = False
        _PluginSettings._DashSupported_Speed_TrueWind_WhenStationary = False ': The wind you feel When stationary Is the True wind.
        _PluginSettings._DashSupported_Speed_VerticalSpeedIndicator_VSI = False

        'Rpm
        _PluginSettings._DashSupported_Rpm = True
        _PluginSettings._DashSupported_RPM_Max = False
        _PluginSettings._DashSupported_Rpm_MaxPercent = False

        'Position
        _PluginSettings._DashSupported_Position_Gear = True
        _PluginSettings._DashSupported_Position_AltitudeIndicator_AI = False '[horizon indicator] artificial horizon
        _PluginSettings._DashSupported_Position_DistanceTraveled = False 'Knotmeter: Measures a boat's speed, and knotlog records distance traveled through the water.
        _PluginSettings._DashSupported_Position_Distance2Destination = False 'Knotmeter: Measures a boat's speed, and knotlog records distance traveled through the water.
        _PluginSettings._DashSupported_Position_DepthFinder = False 'Indicates how deep the water Is so you can avoid running aground

        'Lights
        _PluginSettings._DashSupported_Lights_HeadLights = False
        _PluginSettings._DashSupported_Lights_BreakLights = False
        _PluginSettings._DashSupported_Lights_HazardOrWarning = False
        _PluginSettings._DashSupported_Lights_ShiftIndicator_Up = False
        _PluginSettings._DashSupported_Lights_ShiftIndicator_Down = False
        _PluginSettings._DashSupported_Lights_LeftTurnBlinker = False
        _PluginSettings._DashSupported_Lights_RightTurnBlinker = False
        _PluginSettings._DashSupported_Lights_TaxiLights = False
        _PluginSettings._DashSupported_Lights_TurnOffLights = False
        _PluginSettings._DashSupported_Lights_LandingLights = False
        _PluginSettings._DashSupported_Lights_NavigationLights = False
        _PluginSettings._DashSupported_Lights_AuxileryLights = False

        'Temperature
        _PluginSettings._DashSupported_Temp_OilTemp = False
        _PluginSettings._DashSupported_Temp_CoolantTemp = False
        _PluginSettings._DashSupported_Temp_EngineTemp = False
        _PluginSettings._DashSupported_Temp_CarTrackTemp = False
        _PluginSettings._DashSupported_Temp_PlaneAirTemp = False
        _PluginSettings._DashSupported_Temp_BoatWaterTemp = False
        _PluginSettings._DashSupported_Temp_TrackExitTemp = False

        'Pressure
        _PluginSettings._DashSupported_Pressure_FuelPressure = False
        _PluginSettings._DashSupported_Pressure_OilPressure = False
        _PluginSettings._DashSupported_Pressure_CoolantPressure = False
        _PluginSettings._DashSupported_Pressure_TurboPressure = False

        'Capacity
        _PluginSettings._DashSupported_Capacity_FuelLevel = True
        _PluginSettings._DashSupported_Capacity_FuelCapacity = False
        _PluginSettings._DashSupported_Capacity_BatteryLevel = False
        _PluginSettings._DashSupported_Capacity_BatteryCapacity = False

        'ID
        _PluginSettings._DashSupported_ID_TeamID = False
        _PluginSettings._DashSupported_ID_TrackID = False
        _PluginSettings._DashSupported_ID_VehicleID = False
        _PluginSettings._DashSupported_ID_DriverID = False

        'Lap
        _PluginSettings._DashSupported_Lap_Number = False
        _PluginSettings._DashSupported_Lap_BestTime = False
        _PluginSettings._DashSupported_Lap_PreviousTime = False
        _PluginSettings._DashSupported_Lap_CurrentTime = False
        _PluginSettings._DashSupported_Lap_CurrentPosition = False
        _PluginSettings._DashSupported_Lap_VehiclesinRace = False

        'Orientation
        _PluginSettings._DashSupported_Orient_GPS = False
        _PluginSettings._DashSupported_Orient_Compass = False
        _PluginSettings._DashSupported_Orient_AngleOfAttack_AoA = False
        _PluginSettings._DashSupported_Orient_TurnCoordinator_TC = False 'turn balance indicator
        _PluginSettings._DashSupported_Orient_HeadingIndicator_HI_HSI = False '(HI)Or(HSI)
        _PluginSettings._DashSupported_Orient_WindAngleInstrument = False ': Displays wind angle

        'Alarms
        _PluginSettings._DashSupported_Alarm_EngineOverHeating = False
        _PluginSettings._DashSupported_Alarm_BrakesOverHeating = False
        _PluginSettings._DashSupported_Alarm_WeaponsOverHeating = False
        _PluginSettings._DashSupported_Alarm_StallWarning = False
        _PluginSettings._DashSupported_Alarm_MissleLaunchDetected = False
        _PluginSettings._DashSupported_Alarm_FighterAircraftHaveBoost = False
        _PluginSettings._DashSupported_Alarm_AnchorDragging = False
        _PluginSettings._DashSupported_Alarm_ShallowWater = False

        'Damage
        _PluginSettings._DashSupported_Damage_DamageDetected = False
        _PluginSettings._DashSupported_Damage_ChassisDamage = False
        _PluginSettings._DashSupported_Damage_SuspensionDamage = False
        _PluginSettings._DashSupported_Damage_TireDamage = False
        _PluginSettings._DashSupported_Damage_PartsDetached = False

        'Weapons
        _PluginSettings._DashSupported_Weapons_GunsAmmoCount1 = False
        _PluginSettings._DashSupported_Weapons_GunsAmmoCount2 = False
        _PluginSettings._DashSupported_Weapons_BombsAmmoCount1 = False
        _PluginSettings._DashSupported_Weapons_BombsAmmoCount2 = False
        _PluginSettings._DashSupported_Weapons_CannonsAmmoCount1 = False
        _PluginSettings._DashSupported_Weapons_CannonsAmmoCount2 = False
        _PluginSettings._DashSupported_Weapons_FlairsAmmoCount1 = False
        _PluginSettings._DashSupported_Weapons_FlairsAmmoCount2 = False
        _PluginSettings._DashSupported_Weapons_OtherAmmoCount1 = False
        _PluginSettings._DashSupported_Weapons_OtherAmmoCount2 = False

        'Status
        _PluginSettings._DashSupported_Status_ABS = False
        _PluginSettings._DashSupported_Status_HandBreak = False
        _PluginSettings._DashSupported_Status_InPit = False
        _PluginSettings._DashSupported_Status_PitLimiter = False
        _PluginSettings._DashSupported_Status_AutoPilot = False
        _PluginSettings._DashSupported_Status_LandingGear = False
        _PluginSettings._DashSupported_Status_AirBrakes = False
        _PluginSettings._DashSupported_Status_Flaps = False
        _PluginSettings._DashSupported_Status_ChuteDeployed = False
        _PluginSettings._DashSupported_Status_LandingGearBayDoors = False
        _PluginSettings._DashSupported_Status_BombBayDoors = False
        _PluginSettings._DashSupported_Status_Canopy = False
        _PluginSettings._DashSupported_Status_Trim = False
        _PluginSettings._DashSupported_Status_AnchorUpDown = False
        _PluginSettings._DashSupported_Status_Flags = False
        _PluginSettings._DashSupported_Status_FrontArb = False
        _PluginSettings._DashSupported_Status_RearArb = False
        _PluginSettings._DashSupported_Status_TractionControl = False
        _PluginSettings._DashSupported_Status_Diff = False
        _PluginSettings._DashSupported_Status_BrakeBias = False

        'Extras         
        _PluginSettings._DashSupported_Dash01 = "" '(Add a name to turn it on)
        _PluginSettings._DashSupported_Dash02 = "" '("" = off)
        _PluginSettings._DashSupported_Dash03 = ""
        _PluginSettings._DashSupported_Dash04 = ""
        _PluginSettings._DashSupported_Dash05 = ""
        _PluginSettings._DashSupported_Dash06 = ""
        _PluginSettings._DashSupported_Dash07 = ""
        _PluginSettings._DashSupported_Dash08 = ""
        _PluginSettings._DashSupported_Dash09 = ""
        _PluginSettings._DashSupported_Dash10 = ""
        _PluginSettings._DashSupported_Dash11 = ""
        _PluginSettings._DashSupported_Dash12 = ""
        _PluginSettings._DashSupported_Dash13 = ""
        _PluginSettings._DashSupported_Dash14 = ""
    End Sub

    'Used when the Game Starts
    Public Sub GameStart() Implements IPlugin_Game_v3.GameStart
        'DO SOMETHING HERE AT GAME START!
        'start collecting data
        If GameStarted = False Then
            GameStarted = True
            Dim trd As New Threading.Thread(AddressOf Process_MemoryMap) With {
                .IsBackground = True
            }
            trd.Start()
        End If
    End Sub

    'Used when the Game Stops
    Public Sub GameStop() Implements IPlugin_Game_v3.GameEnd
        'DO SOMETHING HERE AT GAME STOP!
        GameStarted = False
    End Sub

    'Used to inform the user where to Patch the Game 
    Public Function PatchPathInfo() As String Implements IPlugin_Game_v3.PatchPathInfo
        'Tell the User where to patch the game 
        Return ""
    End Function

    'Used to Patch a Game.
    Public Function PatchGame(MyPath As String) As Boolean Implements IPlugin_Game_v3.PatchGame
        Return True
    End Function

    'Used to UnPatch a Game.
    Public Function UnPatchGame(MyPath As String) As Boolean Implements IPlugin_Game_v3.UnPatchGame
        Return True
    End Function

    'Used to Validate the Pathing Path.
    Public Function ValidatePatchPath(MyPath As String) As Boolean Implements IPlugin_Game_v3.ValidatePatchPath
        'insert a simple validation of the patching path - let the user know he got it right
        Return True
    End Function

    '(Optional) Amything extra needed to get the plugin running. (like in game settings)
    Public Function ExtraSetupInfo() As String Implements IPlugin_Game_v3.ExtraSetupInfo
        'Tell the user anything else needed to get the game working properly
        Return ""   'return an empty string to dissable
    End Function

    '//////////////////////////////////////////////////////////////////
    '///   Place The Extra Needed Code For This Plugin Below Here   ///
    '//////////////////////////////////////////////////////////////////
    Private GameStarted As Boolean = False ' our running flag
    Public handle As New GCHandle
    Public AC_Physics As New AC_MemStruct.SPageFilePhysics
    Public AC_Static As New AC_MemStruct.SPageFileStatic
    Public AC_Graphic As New AC_MemStruct.SPageFileGraphic
    Private ReadOnly bytes_Physics(Marshal.SizeOf(AC_Physics)) As Byte
    Private ReadOnly bytes_Static(Marshal.SizeOf(AC_Static)) As Byte
    Private ReadOnly bytes_Graphic(Marshal.SizeOf(AC_Graphic)) As Byte

    'Used by GameManager to Process a MemoryMap.
    Private Sub Process_MemoryMap()
        Dim mmf_physics As MemoryMappedFile = Nothing
        Dim mmf_static As MemoryMappedFile = Nothing
        Dim mmf_graphics As MemoryMappedFile = Nothing
        Dim _physics_connected As Boolean = False
        Dim _static_connected As Boolean = False
        Dim _graphics_connected As Boolean = False

        'loop
        Dim LoopCount As Integer = 0
        Do While GameStarted = True
            'physics
            If _physics_connected = False Then
                Try
                    mmf_physics = MemoryMappedFile.OpenExisting("Local\acpmf_physics")
                    _physics_connected = True
                Catch ex As Exception
                    _physics_connected = False
                End Try
            Else
                ' create acessor
                Using Stream As MemoryMappedViewStream = mmf_physics.CreateViewStream(0, 0, MemoryMappedFileAccess.Read)
                    Stream.Read(bytes_Physics, 0, Marshal.SizeOf(AC_Physics))
                End Using
                handle = GCHandle.Alloc(bytes_Physics, GCHandleType.Pinned)
                AC_Physics = CType(Marshal.PtrToStructure(handle.AddrOfPinnedObject, GetType(AC_MemStruct.SPageFilePhysics)), AC_MemStruct.SPageFilePhysics)
                handle.Free()
                'process
                ProcessPhysics()
            End If

            'static
            If _static_connected = False Then
                Try
                    mmf_static = MemoryMappedFile.OpenExisting("Local\acpmf_static")
                    _static_connected = True
                Catch ex As Exception
                    _static_connected = False
                End Try
            ElseIf LoopCount = 0 Then
                ' create acessor
                Using Stream As MemoryMappedViewStream = mmf_static.CreateViewStream(0, 0, MemoryMappedFileAccess.Read)
                    Stream.Read(bytes_Static, 0, Marshal.SizeOf(AC_Static))
                End Using
                handle = GCHandle.Alloc(bytes_Static, GCHandleType.Pinned)
                AC_Static = CType(Marshal.PtrToStructure(handle.AddrOfPinnedObject, GetType(AC_MemStruct.SPageFileStatic)), AC_MemStruct.SPageFileStatic)
                handle.Free()
                'process
                ProcessStatus()
            End If

            'graphics
            If _graphics_connected = False Then
                Try
                    mmf_graphics = MemoryMappedFile.OpenExisting("Local\acpmf_graphics")
                    _graphics_connected = True
                Catch ex As Exception
                    _graphics_connected = False
                End Try
            ElseIf LoopCount = 0 Then
                ' create acessor
                Using Stream As MemoryMappedViewStream = mmf_graphics.CreateViewStream(0, 0, MemoryMappedFileAccess.Read)
                    Stream.Read(bytes_Graphic, 0, Marshal.SizeOf(AC_Graphic))
                End Using
                handle = GCHandle.Alloc(bytes_Graphic, GCHandleType.Pinned)
                AC_Graphic = CType(Marshal.PtrToStructure(handle.AddrOfPinnedObject, GetType(AC_MemStruct.SPageFileGraphic)), AC_MemStruct.SPageFileGraphic)
                handle.Free()
                'process
                ProcessGraphics()
            End If

            'add one
            If LoopCount = 49 Then 'max loop count - how fast static and graphics collects data ~250ms at 49 (5ms * 50)
                LoopCount = 0
            Else
                LoopCount += 1
            End If

            'Sleep - 5 millisecond input
            Threading.Thread.Sleep(5)
        Loop

        'close physics
        Try
            mmf_physics.Dispose()
        Catch ex As Exception
        End Try

        'close status
        Try
            mmf_static.Dispose()
        Catch ex As Exception
        End Try

        'close graphics
        Try
            mmf_graphics.Dispose()
        Catch ex As Exception
        End Try
    End Sub

    'Process Physics 
    Private Sub ProcessPhysics()
        With AC_Physics
            'assign values - clip very small values
            Dim RadtoDeg As Double = 45.0 / Atan(1)

            'Motion
            _Motion_Outputs._Roll = - .roll * RadtoDeg
            _Motion_Outputs._Pitch = - .pitch * RadtoDeg
            _Motion_Outputs._Yaw = - .heading * RadtoDeg
            _Motion_Outputs._Heave = .accG.Y
            _Motion_Outputs._Sway = - .accG.X
            _Motion_Outputs._Surge = .accG.Z
            _Motion_Outputs._Extra1 = .LocalVelocity.X
            _Motion_Outputs._Extra4 = .LocalVelocity.X
            _Motion_Outputs._Extra5 = .LocalVelocity.Z
            _Motion_Outputs._Extra6 = .LocalVelocity.Y

            'Dash
            _Dash_Outputs._Speed_GroundSpeed = .speedKmh.ToString
            _Dash_Outputs._Rpm = Int(.rpms + 0.5).ToString
            _Dash_Outputs._Position_Gear = .gear.ToString
            _Dash_Outputs._Capacity_FuelLevel = .fuel.ToString

            'Vibe
            _Vibe_Outputs._Vibe01 = Int(.rpms + 0.5)
            _Vibe_Outputs._Vibe02 = .gear
            _Vibe_Outputs._Vibe03 = _Motion_Outputs._Sway
            _Vibe_Outputs._Vibe04 = _Motion_Outputs._Surge
            _Vibe_Outputs._Vibe05 = _Motion_Outputs._Heave
            _Vibe_Outputs._Vibe_Chassis_FL = .suspensionTravel.Front_Left * 10000
            _Vibe_Outputs._Vibe_Chassis_FR = .suspensionTravel.Front_Right * 10000
            _Vibe_Outputs._Vibe_Chassis_RL = .suspensionTravel.Rear_Left * 10000
            _Vibe_Outputs._Vibe_Chassis_RR = .suspensionTravel.Rear_Right * 10000
        End With
    End Sub

    'Process Status
    Private Sub ProcessStatus()
        With AC_Static
            _Status_Outputs._CurrentVehicle = .carModel
        End With
    End Sub

    'Process Graphics
    Private Sub ProcessGraphics()
        With AC_Graphic
            If .status = AC_MemStruct.AC_STATUS.AC_LIVE OrElse .status = AC_MemStruct.AC_STATUS.AC_REPLAY Then
                _Status_Outputs._OutputActive = True
            Else
                _Status_Outputs._OutputActive = False
            End If
        End With
    End Sub
End Class